// This file is automatically generated, do not modify!
EXPORT_FUNCTION(check_all_files_crc32_info)
EXPORT_FUNCTION(clear_all_caches)
EXPORT_FUNCTION(compute_crc32)
EXPORT_FUNCTION(displayln_malloc_stats)
EXPORT_FUNCTION(does_file_exist)
EXPORT_FUNCTION(does_file_exist_qar)
EXPORT_FUNCTION(does_regular_file_exist_qar)
EXPORT_FUNCTION(free_lat_data)
EXPORT_FUNCTION(get_all_caches_info)
EXPORT_FUNCTION(get_dim_indices_lat_data)
EXPORT_FUNCTION(get_dim_name_lat_data)
EXPORT_FUNCTION(get_dim_size_lat_data)
EXPORT_FUNCTION(get_dim_sizes_lat_data)
EXPORT_FUNCTION(get_ndim_lat_data)
EXPORT_FUNCTION(get_qar_multi_vol_max_size)
EXPORT_FUNCTION(is_complex_lat_data)
EXPORT_FUNCTION(is_directory)
EXPORT_FUNCTION(is_matching_lat_data)
EXPORT_FUNCTION(is_regular_file)
EXPORT_FUNCTION(list_qar)
EXPORT_FUNCTION(load_lat_data)
EXPORT_FUNCTION(mk_lat_data)
EXPORT_FUNCTION(peek_lat_data)
EXPORT_FUNCTION(poke_lat_data)
EXPORT_FUNCTION(qappend)
EXPORT_FUNCTION(qappend_info)
EXPORT_FUNCTION(qar_create)
EXPORT_FUNCTION(qar_extract)
EXPORT_FUNCTION(qcat)
EXPORT_FUNCTION(qcat_bytes)
EXPORT_FUNCTION(qcopy_file)
EXPORT_FUNCTION(qload_datatable)
EXPORT_FUNCTION(qls)
EXPORT_FUNCTION(qls_all)
EXPORT_FUNCTION(qmkdir)
EXPORT_FUNCTION(qmkdir_info)
EXPORT_FUNCTION(qnorm_lat_data)
EXPORT_FUNCTION(qremove)
EXPORT_FUNCTION(qremove_all)
EXPORT_FUNCTION(qrename)
EXPORT_FUNCTION(qrename_info)
EXPORT_FUNCTION(qtouch)
EXPORT_FUNCTION(qtouch_info)
EXPORT_FUNCTION(random_permute)
EXPORT_FUNCTION(save_lat_data)
EXPORT_FUNCTION(set_add_lat_data)
EXPORT_FUNCTION(set_dim_name_lat_data)
EXPORT_FUNCTION(set_dim_sizes_lat_data)
EXPORT_FUNCTION(set_lat_data)
EXPORT_FUNCTION(set_mul_double_lat_data)
EXPORT_FUNCTION(set_sub_lat_data)
EXPORT_FUNCTION(set_zero_lat_data)
EXPORT_FUNCTION(show_lat_data)
