cimport libcpp.complex
from libc.stdint cimport uint64_t
from libc.stdint cimport int64_t
from libc.stdint cimport uint8_t
from libc.stdint cimport int8_t

ctypedef double complex Complex
# ctypedef libcpp.complex.complex[double] Complex

ctypedef float complex ComplexF
# ctypedef libcpp.complex.complex[float] Complex

ctypedef double Double

ctypedef float Float

ctypedef long Long

ctypedef int64_t Int64t

ctypedef int8_t Int8t

ctypedef char Char
